from rest_framework.utils.html import *  # NOQA

# This is a verbatim "backport" of a pending patch to Django REST Framework.
# See https://github.com/tomchristie/django-rest-framework/pull/2682.


def parse_json_form(dictionary, prefix=''):
    """
    Parse an HTML JSON form submission as per the W3C Draft spec
    An implementation of "The application/json encoding algorithm"
    http://www.w3.org/TR/html-json-forms/
    """
    # Step 1: Initialize output object
    output = {}
    for name, value in get_all_items(dictionary):
        # TODO: implement is_file flag

        # Step 2: Compute steps array
        steps = parse_json_path(name)

        # Step 3: Initialize context
        context = output

        # Step 4: Iterate through steps
        for step in steps:
            # Step 4.1 Retrieve current value from context
            current_value = get_value(context, step.key, Undefined())

            # Steps 4.2, 4.3: Set JSON value on context
            context = set_json_value(
                context=context,
                step=step,
                current_value=current_value,
                entry_value=value,
                is_file=False,
            )
    # Convert any remaining Undefined array entries to None
    output = clean_undefined(output)
    output = clean_empty_string(output)

    # Account for DRF prefix (not part of JSON form spec)
    result = get_value(output, prefix, Undefined())
    if isinstance(result, Undefined):
        return output
    else:
        return result


def parse_json_path(path):
    """
    Parse a string as a JSON path
    An implementation of "steps to parse a JSON encoding path"
    http://www.w3.org/TR/html-json-forms/#dfn-steps-to-parse-a-json-encoding-path
    """

    # Steps 1, 2, 3
    original_path = path
    steps = []

    # Step 11 (Failure)
    failed = [
        JsonStep(
            type="object",
            key=original_path,
            last=True,
            failed=True,
        )
    ]

    # Other variables for later use
    digit_re = re.compile(r'^\[([0-9]+)\]')
    key_re = re.compile(r'^\[([^\]]+)\]')

    # Step 4 - Find characters before first [ (if any)
    parts = path.split("[")
    first_key = parts[0]
    if parts[1:]:
        path = "[" + "[".join(parts[1:])
    else:
        path = ""

    # Step 5 - According to spec, keys cannot start with [
    # NOTE: This was allowed in older DRF versions, so disabling rule for now
    # if not first_key:
    #     return failed

    # Step 6 - Save initial step
    steps.append(JsonStep(
        type="object",
        key=first_key,
    ))

    # Step 7 - Simple single-step case (no [ found)
    if not path:
        steps[-1].last = True
        return steps

    # Step 8 - Loop
    while path:
        # Step 8.1 - Check for single-item array
        if path[:2] == "[]":
            steps[-1].append = True
            path = path[2:]
            if path:
                return failed
            continue

        # Step 8.2 - Check for array[index]
        digit_match = digit_re.match(path)
        if digit_match:
            path = digit_re.sub("", path)
            steps.append(JsonStep(
                type="array",
                key=int(digit_match.group(1)),
            ))
            continue

        # Step 8.3 - Check for object[key]
        key_match = key_re.match(path)
        if key_match:
            path = key_re.sub("", path)
            steps.append(JsonStep(
                type="object",
                key=key_match.group(1),
            ))
            continue

        # Step 8.4 - Invalid key format
        return failed

    # Step 9
    next_step = None
    for step in reversed(steps):
        if next_step:
            step.next_type = next_step.type
        else:
            step.last = True
        next_step = step

    return steps


def set_json_value(context, step, current_value, entry_value, is_file):
    """
    Apply a JSON value to a context object
    An implementation of "steps to set a JSON encoding value"
    http://www.w3.org/TR/html-json-forms/#dfn-steps-to-set-a-json-encoding-value
    """

    # TODO: handle is_file

    # Add empty values to array so indexing works like JavaScript
    if isinstance(context, list) and isinstance(step.key, int):
        while len(context) <= step.key:
            context.append(Undefined())

    # Step 7: Handle last step
    if step.last:
        if isinstance(current_value, Undefined):
            # Step 7.1: No existing value
            key = step.key
            if isinstance(context, dict) and isinstance(key, int):
                key = str(key)
            if step.append:
                context[key] = [entry_value]
            else:
                context[key] = entry_value
        elif isinstance(current_value, list):
            # Step 7.2: Existing value is an Array, assume multi-valued field
            # and add entry to end.

            # FIXME: What if the other items in the array had explicit keys and
            # this one is supposed to be the "" value?
            # (See step 8.4 and Example 7)
            context[step.key].append(entry_value)

        elif isinstance(current_value, dict) and not is_file:
            # Step 7.3: Existing value is an Object
            return set_json_value(
                context=current_value,
                step=JsonStep(type="object", key="", last=True),
                current_value=current_value.get("", Undefined()),
                entry_value=entry_value,
                is_file=is_file,
            )
        else:
            # Step 7.4: Existing value is a scalar; preserve both values
            context[step.key] = [current_value, entry_value]

        # Step 7.5
        return context

    # Step 8: Handle intermediate steps
    if isinstance(current_value, Undefined):
        # 8.1: No existing value
        if step.next_type == "array":
            context[step.key] = []
        else:
            context[step.key] = {}
        return context[step.key]
    elif isinstance(current_value, dict):
        # Step 8.2: Existing value is an Object
        return get_value(context, step.key, Undefined())
    elif isinstance(current_value, list):
        # Step 8.3: Existing value is an Array
        if step.next_type == "array":
            return current_value
        # Convert array to object to facilitate mixed keys
        obj = {}
        for i, item in enumerate(current_value):
            if not isinstance(item, Undefined):
                obj[str(i)] = item
        context[step.key] = obj
        return obj
    else:
        # 8.4: Existing value is a scalar; convert to Object, preserving
        # current value via an empty key
        obj = {'': current_value}
        context[step.key] = obj
        return obj


def get_value(obj, key, default=None):
    """
    Mimic JavaScript Object/Array behavior by allowing access to nonexistent
    indexes.
    """
    if isinstance(obj, dict):
        return obj.get(key, default)
    elif isinstance(obj, list):
        try:
            return obj[key]
        except IndexError:
            return default


class Undefined(object):
    """
    Use Undefined() rather than None to distinguish from null.
    """
    pass


class JsonStep(object):
    """
    Struct to represent "step" as described in HTML JSON form algorithm
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        vals = ",".join(
            "%s=%s" % (key, val) for key, val in self.__dict__.items()
        )
        return "JsonStep(%s)" % vals

    type = None
    next_type = None
    key = None
    append = None
    last = None
    failed = None


def clean_undefined(obj):
    """
    Convert Undefined array entries to None (null)
    """
    if isinstance(obj, list):
        return [
            None if isinstance(item, Undefined) else item
            for item in obj
        ]
    if isinstance(obj, dict):
        for key in obj:
            obj[key] = clean_undefined(obj[key])
    return obj


def clean_empty_string(obj):
    """
    Replace empty form values with None, since the is_html_input() check in
    Field won't work after we convert to JSON.
    (FIXME: What about allow_blank=True?)
    """
    if obj == '':
        return None
    if isinstance(obj, list):
        return [
            None if item == '' else item
            for item in obj
        ]
    if isinstance(obj, dict):
        for key in obj:
            obj[key] = clean_empty_string(obj[key])
    return obj


def get_all_items(obj):
    """
    dict.items() but with a separate row for each value in a MultiValueDict
    """
    if hasattr(obj, 'getlist'):
        items = []
        for key in obj:
            for value in obj.getlist(key):
                items.append((key, value))
        return items
    else:
        return obj.items()



import warnings
warnings.warn(
    "The wq.db.rest.app API has been moved to wq.db.rest. "
    "The actual Router implementation has moved to wq.db.rest.routers",
    DeprecationWarning
)


from . import ModelRouter as Router, router, autodiscover  # NOQA



from django.contrib import admin
from django.contrib.contenttypes.admin import GenericTabularInline

from .models import Annotation, AnnotatedModel
from .forms import AnnotationForm


class AnnotationInline(GenericTabularInline):
    model = Annotation
    form = AnnotationForm
    extra = 0


class AnnotatedModelAdmin(admin.ModelAdmin):
    model = AnnotatedModel
    inlines = [
        AnnotationInline,
    ]


class AnnotationTypeAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'contenttype')
    list_filter = ('contenttype',)



from rest_framework import serializers
from wq.db.patterns.base import serializers as base
from wq.db.rest.models import get_object_id
from .models import Identifier, Authority


class IdentifierListSerializer(base.TypedAttachmentListSerializer):
    def to_internal_value(self, data):
        data = super(IdentifierListSerializer, self).to_internal_value(data)
        primary = [
            ident for ident in data
            if ident and ident.get('is_primary', None)
        ]
        if not any(primary) and len(data) > 0:
            for ident in data:
                if ident:
                    ident['is_primary'] = True
                    break
        return data


class IdentifierSerializer(base.TypedAttachmentSerializer):
    type_model = Authority
    type_field = 'authority_id'
    attachment_fields = ['id', 'name', 'slug', 'is_primary']
    required_fields = ['name']

    url = serializers.ReadOnlyField()

    @property
    def expected_types(self):
        return [None] + list(Authority.objects.all())

    class Meta(base.TypedAttachmentSerializer.Meta):
        model = Identifier
        list_serializer_class = IdentifierListSerializer


class IdentifiedModelSerializer(base.AttachedModelSerializer):
    id = serializers.SerializerMethodField()
    identifiers = IdentifierSerializer(many=True)

    def get_id(self, instance):
        return get_object_id(instance)

    class Meta:
        list_exclude = ('identifiers',)



from rest_framework.filters import BaseFilterBackend
RESERVED_PARAMETERS = ('_', 'page', 'limit', 'format', 'slug', 'mode')

from .models import get_ct, get_by_identifier
from django.utils.six import string_types
from django.db.models.fields import FieldDoesNotExist


class FilterBackend(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        kwargs = {}
        for key, val in list(view.kwargs.items()) + list(request.GET.items()):
            if key in RESERVED_PARAMETERS or key in view.ignore_kwargs:
                continue
            kwargs[key] = val if isinstance(val, string_types) else val[0]
        model = getattr(view, 'model', None) or queryset.model
        ctype = get_ct(model)

        for key, val in list(kwargs.items()):
            if key.startswith('related_'):
                continue
            field_name = key.split('__')[0]
            try:
                field = model._meta.get_field_by_name(field_name)[0]
            except FieldDoesNotExist:
                del kwargs[key]
                continue

            if field_name != key:
                continue

            rel = getattr(field, 'rel', None)
            if not rel:
                continue

            pcls = field.rel.to
            router = getattr(view, 'router', None)
            if router:
                slug = router.get_lookup_for_model(pcls)
                kwargs[key] = pcls.objects.get(**{slug: val})
            elif get_ct(pcls).is_identified:
                kwargs[key] = pcls.objects.get_by_identifier(val)
            else:
                kwargs[key] = pcls.objects.get(pk=val)

        for key, val in list(kwargs.items()):
            if key.startswith('related_') and ctype.is_related:
                for pct in ctype.get_all_parents():
                    if key == 'related_' + pct.identifier:
                        pclass = pct.model_class()
                        parent = get_by_identifier(pclass.objects, kwargs[key])
                        del kwargs[key]
                        objs = model.objects.filter_by_related(parent)
                        kwargs['pk__in'] = objs.values_list('pk', flat=True)

        return queryset.filter(**kwargs)



from rest_framework import serializers
from wq.db.rest.serializers import ModelSerializer
from wq.db.patterns.base import serializers as base
from .models import FileField, FileType
from django.core.files.uploadedfile import UploadedFile


class FileSerializer(ModelSerializer):
    is_image = serializers.ReadOnlyField()

    def __init__(self, *args, **kwargs):
        self.serializer_field_mapping[FileField] = serializers.FileField
        super(FileSerializer, self).__init__(*args, **kwargs)

    def to_internal_value(self, data):
        obj = super(FileSerializer, self).to_internal_value(data)
        if obj and hasattr(obj, 'user') and obj.user is None:
            if 'request' in self.context:
                user = self.context['request'].user
                if user.is_authenticated():
                    obj.user_id = user.pk
        return obj


class FileAttachmentListSerializer(base.TypedAttachmentListSerializer):
    def get_value(self, dictionary):
        values = dictionary.get(self.source, None)
        if not isinstance(values, list):
            values = [values]
        if all(isinstance(value, UploadedFile) for value in values):
            return [
                {'file': value} for value in values
            ]
        return super(FileAttachmentListSerializer, self).get_value(dictionary)


class FileAttachmentSerializer(base.TypedAttachmentSerializer, FileSerializer):
    attachment_fields = ['id', 'name', 'file']
    type_model = FileType

    class Meta(base.TypedAttachmentSerializer.Meta):
        list_serializer_class = FileAttachmentListSerializer


class FileAttachedModelSerializer(base.AttachedModelSerializer):
    pass



from wq.db.rest.serializers import ModelSerializer
from rest_framework import serializers
from rest_framework.utils import model_meta
from wq.db.rest.models import get_ct, get_object_id
from wq.db.rest import compat as html
from .models import NaturalKeyModel


class AttachmentListSerializer(serializers.ListSerializer):
    def to_representation(self, data):
        data = super(AttachmentListSerializer, self).to_representation(data)
        if self.parent:
            for i, row in enumerate(data):
                row['@index'] = i
        return data


class TypedAttachmentListSerializer(AttachmentListSerializer):
    def get_value(self, dictionary):
        # Handle attachments that are submitted together with their parent

        # Ideal case: an array of dicts; this can be handled by the default
        # implementation.  HTML JSON forms will use this approach.
        if self.field_name in dictionary:
            value = super(TypedAttachmentListSerializer, self).get_value(
                dictionary
            )
            for i, row in enumerate(value):
                empty = True
                for key, val in row.items():
                    if key == self.child.type_field:
                        continue
                    elif val:
                        empty = False
                if empty:
                    value[i] = None
            return value

        # Deprecated/"classic" form style, where each attachment is submitted
        # as form fields with names in the format [model]_[typeid] e.g.
        # annotation_23=30.5 will become {'type': 23, 'value': '30.5'}

        # Retrieve form values into more usable array/dict format
        attachments = []
        ct = get_ct(self.child.Meta.model)
        i = 0
        for atype in self.child.expected_types:
            fields = {
                afield: '%s-%s-%s' % (
                    ct.identifier,
                    get_object_id(atype) if atype else '',
                    afield
                ) for afield in self.child.attachment_fields
            }
            found = False
            for key in fields.values():
                if key in dictionary:
                    found = True
            if found:
                attachment = self.child.create_dict(
                    atype, dictionary, fields, i
                )
                if attachment:
                    attachments.append(attachment)
                i += 1

        # Return extracted values to default implementation
        return attachments


class AttachmentSerializer(ModelSerializer):
    id = serializers.IntegerField(required=False)

    def __init__(self, *args, **kwargs):
        kwargs['allow_null'] = True
        super(AttachmentSerializer, self).__init__(*args, **kwargs)


class TypedAttachmentSerializer(AttachmentSerializer):
    attachment_fields = ['id']
    required_fields = []
    type_model = None
    type_field = 'type_id'
    object_field = 'content_object'

    def create_dict(self, atype, data, fields, index):
        attachment = {
            self.type_field: atype.pk if atype else None,
        }
        for field in fields:
            if fields[field] in data:
                attachment[field] = data[fields[field]]
                if field == 'id' and attachment[field]:
                    attachment[field] = int(attachment[field])
        for field in self.required_fields:
            if not attachment.get(field, None):
                return None
        return attachment

    @property
    def expected_types(self):
        return self.type_model.objects.all()

    def to_representation(self, obj):
        data = super(TypedAttachmentSerializer, self).to_representation(obj)

        has_parent = (
            self.parent and self.parent.parent
            and hasattr(self.parent.parent.Meta, 'model')
        )
        if has_parent:
            # This is being serialized with its parent object, don't need to
            # reference the parent again.
            pass
        else:
            # Include pointer to parent object. Collapse content_type and
            # object_id into a single attribute: the identifier for the content
            # type plus '_id' e.g. if contenttype's name is 'visit' then the
            # attribute is 'visit_id'.  This lets us pretend the generic
            # foreign key is a regular one in client.  For the value, Use
            # get_object_id insted of obj.object_id, in case the the content
            # object is an IdentifiedModel

            parent_obj = getattr(obj, self.object_field)
            idname = get_ct(parent_obj).identifier
            data[idname + '_id'] = get_object_id(parent_obj)

            # In detail views, include full parent object (without _id suffix)
            if self.is_detail:
                from wq.db import rest
                data[idname] = rest.router.serialize(parent_obj)
        return data

    class Meta:
        # Don't validate these fields (items are saved with their parent)
        exclude = ("content_type", "object_id",)
        list_serializer_class = TypedAttachmentListSerializer


class AttachedModelSerializer(ModelSerializer):
    def create(self, validated_data):
        model_data, attachment_data = self.extract_attachments(validated_data)
        instance = super(AttachedModelSerializer, self).create(model_data)

        fields = self.get_fields()
        for name in attachment_data:
            model = fields[name].child.Meta.model
            for attachment in attachment_data[name]:
                if not attachment:
                    continue
                self.set_parent_object(attachment, instance, name)
                self.create_attachment(model, attachment, name)
        return instance

    def update(self, instance, validated_data):
        model_data, attachment_data = self.extract_attachments(validated_data)
        obj = super(
            AttachedModelSerializer, self
        ).update(instance, model_data)

        fields = self.get_fields()
        for name in attachment_data:
            model = fields[name].child.Meta.model
            for attachment in attachment_data[name]:
                if not attachment:
                    continue
                self.set_parent_object(attachment, instance, name)
                if 'id' in attachment:
                    exist = self.get_attachment(model, attachment['id'])
                    self.update_attachment(exist, attachment, name)
                else:
                    self.create_attachment(model, attachment, name)
        return obj

    def extract_attachments(self, validated_data):
        fields = self.get_fields()
        attachment_data = {}
        for name, field in fields.items():
            if isinstance(field, AttachmentListSerializer):
                attachment_data[name] = validated_data.pop(name, [])
        return validated_data, attachment_data

    def set_parent_object(self, attachment, instance, name):
        serializer = self.get_fields()[name].child
        fk_name = getattr(serializer, 'object_field', 'content_object')
        attachment[fk_name] = instance

    def get_attachment(self, model, pk):
        return model.objects.get(pk=pk)

    def update_attachment(self, exist, attachment, name):
        field = self.get_fields()[name]
        attachment.pop('id')
        field.child.update(exist, attachment)

    def create_attachment(self, model, attachment, name):
        field = self.get_fields()[name]
        field.child.create(attachment)


class NaturalKeyValidator(serializers.UniqueTogetherValidator):
    def set_context(self, serializer):
        self.nested_fields = {
            name: serializer.fields[name]
            for name in self.fields
            if isinstance(serializer.fields[name], NaturalKeySerializer)
        }
        super(NaturalKeyValidator, self).set_context(serializer)

    def filter_queryset(self, attrs, queryset):
        attrs = attrs.copy()
        for field in attrs:
            if field in self.nested_fields:
                assert(isinstance(attrs[field], dict))
                cls = self.nested_fields[field].Meta.model
                result = cls._default_manager.filter(
                    **attrs[field]
                )
                if result.count() == 0:
                    # No existing nested object for these values
                    return queryset.none()
                else:
                    # Existing nested object, use it to validate
                    attrs[field] = result[0].pk

        return super(NaturalKeyValidator, self).filter_queryset(
            attrs, queryset
        )


class NaturalKeySerializer(serializers.ModelSerializer):
    """
    Self-nesting Serializer for NaturalKeyModels
    """
    def build_nested_field(self, field_name, relation_info, nested_depth):
        field_class = NaturalKeySerializer.for_model(
            relation_info.related_model,
            validate_key=False,
        )
        field_kwargs = {}
        return field_class, field_kwargs

    def create(self, validated_data):
        model_class = self.Meta.model
        natural_key_fields = model_class.get_natural_key_fields()
        natural_key = []
        for field in natural_key_fields:
            val = validated_data
            for key in field.split('__'):
                val = val[key]
            natural_key.append(val)
        return model_class.objects.find(*natural_key)

    def update(self, instance, validated_data):
        raise NotImplementedError(
            "Updating an existing natural key is not supported."
        )

    @classmethod
    def for_model(cls, model_class, validate_key=True):
        unique_together = model_class._meta.unique_together[0]

        class Serializer(cls):
            class Meta(cls.Meta):
                model = model_class
                fields = unique_together
                if validate_key:
                    validators = [
                        NaturalKeyValidator(
                            queryset=model_class._default_manager,
                            fields=unique_together,
                        )
                    ]
                else:
                    validators = []
        return Serializer

    @classmethod
    def for_depth(cls, model_class):
        return cls

    def to_internal_value(self, data):
        if html.is_html_input(data):
            data = html.parse_json_form(data)
        result = super(NaturalKeySerializer, self).to_internal_value(data)
        return result

    class Meta:
        depth = 1


class NaturalKeyModelSerializer(AttachedModelSerializer):
    """
    Serializer for models with one or more foreign keys to a NaturalKeyModel
    """
    def build_nested_field(self, field_name, relation_info, nested_depth):
        if issubclass(relation_info.related_model, NaturalKeyModel):
            field_class = NaturalKeySerializer.for_model(
                relation_info.related_model,
                validate_key=False,
            )
            field_kwargs = {}
            if relation_info.model_field.null:
                field_kwargs['required'] = False
            return field_class, field_kwargs

        return super(NaturalKeyModelSerializer, self).build_nested_field(
            field_name, relation_info, nested_depth
        )

    def build_relational_field(self, field_name, relation_info):
        field_class, field_kwargs = super(
            NaturalKeyModelSerializer, self
        ).build_relational_field(
            field_name, relation_info
        )
        if issubclass(relation_info.related_model, NaturalKeyModel):
            field_kwargs.pop('queryset')
            field_kwargs['read_only'] = True
        return field_class, field_kwargs

    def get_fields(self):
        fields = super(NaturalKeyModelSerializer, self).get_fields()
        info = model_meta.get_field_info(self.Meta.model)
        for key in fields:
            if not key.endswith('_id'):
                continue
            field = key[:-3]
            if field in fields or field not in info.relations:
                continue
            relation_info = info.relations[field]
            if not issubclass(relation_info.related_model, NaturalKeyModel):
                continue
            field_class, field_kwargs = self.build_nested_field(
                field, relation_info, 1
            )
            fields[field] = field_class(**field_kwargs)
        return fields

    def create(self, validated_data):
        self.convert_natural_keys(
            validated_data
        )
        return super(NaturalKeyModelSerializer, self).create(
            validated_data
        )

    def update(self, instance, validated_data):
        self.convert_natural_keys(
            validated_data
        )
        return super(NaturalKeyModelSerializer, self).update(
            instance, validated_data
        )

    def convert_natural_keys(self, validated_data):
        fields = self.get_fields()
        for name, field in fields.items():
            if name not in validated_data:
                continue
            if isinstance(field, NaturalKeySerializer):
                validated_data[name] = fields[name].create(
                    validated_data[name]
                )



from django.conf.urls import patterns, url
from rest_framework.urlpatterns import format_suffix_patterns

ids = r'^(?P<ids>[^\.]+)'


def make_urls(views):
    urls = [
        url(ids + r'/' + name + '$', cls.as_view())
        for name, cls in views.items()
    ]
    return format_suffix_patterns(patterns('', *urls))



from wq.db import rest
from .models import Markdown
from .serializers import MarkdownSerializer

import swapper
MarkdownType = swapper.load_model("mark", "MarkdownType")


def active_markdown(qs, request):
    mtype = MarkdownType.get_current(request)
    return qs.filter(type=mtype)

rest.router.register_model(
    Markdown,
    serializer=MarkdownSerializer,
    filter=active_markdown,
    url='markdown',
)
rest.router.register_model(MarkdownType)



from wq.db.patterns.base import serializers as base
from .models import AnnotationType, Annotation


class AnnotationSerializer(base.TypedAttachmentSerializer):
    attachment_fields = ['id', 'value']
    type_model = AnnotationType

    class Meta(base.TypedAttachmentSerializer.Meta):
        model = Annotation


class AnnotatedModelSerializer(base.AttachedModelSerializer):
    annotations = AnnotationSerializer(many=True)

    class Meta:
        list_exclude = ('annotations',)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import swapper


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
        swapper.dependency('mark', 'MarkdownType'),
    ]

    operations = [
        migrations.CreateModel(
            name='MarkdownType',
            fields=[
                ('id', models.AutoField(serialize=False, verbose_name='ID', primary_key=True, auto_created=True)),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'swappable': swapper.swappable_setting('mark', 'MarkdownType'),
                'db_table': 'wq_markdowntype',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Markdown',
            fields=[
                ('id', models.AutoField(serialize=False, verbose_name='ID', primary_key=True, auto_created=True)),
                ('markdown', models.TextField(blank=True, null=True)),
                ('summary', models.CharField(blank=True, max_length=255, null=True)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
                ('type', models.ForeignKey(to=swapper.get_model_name('mark', 'MarkdownType'))),
            ],
            options={
                'db_table': 'wq_markdown',
            },
            bases=(models.Model,),
        ),
    ]



from .base.serializers import *  # NOQA
from .annotate.serializers import *  # NOQA
from .identify.serializers import *  # NOQA
from .locate.serializers import *  # NOQA
from .mark.serializers import *  # NOQA
from .relate.serializers import *  # NOQA


class IdentifiedLocatedModelSerializer(
        IdentifiedModelSerializer, LocatedModelSerializer):
    pass


class IdentifiedMarkedModelSerializer(
        IdentifiedModelSerializer, MarkedModelSerializer):
    class Meta:
        list_exclude = (
            'identifiers',
            'markdown',
        )


class IdentifiedRelatedModelSerializer(
        IdentifiedModelSerializer, RelatedModelSerializer):
    class Meta:
        list_exclude = (
            'identifiers',
            'relationships',
            'inverserelationships',
        )



from django.contrib import admin
from django.contrib.contenttypes.admin import GenericTabularInline
from django.template.base import TemplateDoesNotExist

from .models import Markdown

try:
    from django_markdown.admin import MarkdownInlineAdmin as InlineAdmin
except (ImportError, TemplateDoesNotExist):
    from django.contrib.admin import StackedInline as InlineAdmin


class MarkdownInline(GenericTabularInline, InlineAdmin):
    model = Markdown
    extra = 0


class MarkedModelAdmin(admin.ModelAdmin):
    inlines = [
        MarkdownInline
    ]



from .annotate.admin import *  # NOQA
from .identify.admin import *  # NOQA
from .locate.admin import *  # NOQA
from .mark.admin import *  # NOQA
from .relate.admin import *  # NOQA

from .models import IdentifiedRelatedModel, IdentifiedMarkedModel


class IdentifiedRelatedModelAdmin(IdentifiedModelAdmin, RelatedModelAdmin):
    model = IdentifiedRelatedModel
    inlines = [
        IdentifierInline,
        RelationshipInline,
        InverseRelationshipInline
    ]


class IdentifiedMarkedModelAdmin(IdentifiedModelAdmin, MarkedModelAdmin):
    model = IdentifiedMarkedModel
    inlines = [
        IdentifierInline,
        MarkdownInline,
    ]



from wq.db import rest
from .serializers import AnnotationSerializer
from .models import Annotation, AnnotationType

rest.router.register_model(Annotation, serializer=AnnotationSerializer)
rest.router.register_model(AnnotationType)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings
import wq.db.contrib.files.models

import swapper


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),
                ('name', models.CharField(max_length=255, blank=True, null=True)),
                ('file', wq.db.contrib.files.models.FileField(upload_to='.', height_field='height', width_field='width')),
                ('size', models.IntegerField(blank=True, null=True)),
                ('width', models.IntegerField(blank=True, null=True)),
                ('height', models.IntegerField(blank=True, null=True)),
            ],
            options={
                'swappable': swapper.swappable_setting('files', 'File'),
                'db_table': 'wq_file',
                'ordering': ('name',),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='FileType',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('mimetype', models.CharField(max_length=255)),
            ],
            options={
                'db_table': 'wq_filetype',
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='file',
            name='type',
            field=models.ForeignKey(to='files.FileType', null=True, blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='file',
            name='user',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL, null=True, blank=True),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Relationship',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False, verbose_name='ID', auto_created=True)),
                ('from_object_id', models.PositiveIntegerField()),
                ('to_object_id', models.PositiveIntegerField()),
                ('computed', models.BooleanField(default=False)),
                ('from_content_type', models.ForeignKey(to='contenttypes.ContentType', related_name='+')),
                ('to_content_type', models.ForeignKey(to='contenttypes.ContentType', related_name='+')),
            ],
            options={
                'db_table': 'wq_relationship',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='RelationshipType',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False, verbose_name='ID', auto_created=True)),
                ('name', models.CharField(max_length=255)),
                ('inverse_name', models.CharField(blank=True, null=True, max_length=255)),
                ('computed', models.BooleanField(default=False)),
                ('from_type', models.ForeignKey(to='contenttypes.ContentType', related_name='+')),
                ('to_type', models.ForeignKey(to='contenttypes.ContentType', related_name='+')),
            ],
            options={
                'db_table': 'wq_relationshiptype',
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='relationship',
            name='type',
            field=models.ForeignKey(to='relate.RelationshipType'),
            preserve_default=True,
        ),
        migrations.CreateModel(
            name='InverseRelationship',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('relate.relationship',),
        ),
        migrations.CreateModel(
            name='InverseRelationshipType',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('relate.relationshiptype',),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Authority',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('homepage', models.URLField(null=True, blank=True)),
                ('object_url', models.URLField(null=True, blank=True)),
            ],
            options={
                'db_table': 'wq_identifiertype',
                'verbose_name_plural': 'authorities',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Identifier',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('name', models.CharField(max_length=255, db_index=True)),
                ('slug', models.SlugField(blank=True)),
                ('is_primary', models.BooleanField(default=False)),
                ('object_id', models.PositiveIntegerField()),
                ('authority', models.ForeignKey(null=True, to='identify.Authority', blank=True)),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
            options={
                'db_table': 'wq_identifier',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PrimaryIdentifier',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('identify.identifier',),
        ),
    ]



from rest_framework import serializers
from django.contrib.gis.db import models as model_fields
from django.contrib.gis.geos import GEOSGeometry
from django.utils import timezone

from django.conf import settings

from .models import ContentType, get_ct, get_object_id, get_by_identifier

from rest_framework.utils import model_meta
from . import compat as html


class GeometryField(serializers.Field):
    def to_representation(self, value):
        if value is None:
            return None
        import json
        return json.loads(value.geojson)

    def to_internal_value(self, value):
        import json
        if isinstance(value, dict):
            value = json.dumps(value)
        geom = GEOSGeometry(value)
        srid = getattr(settings, 'SRID', 4326)

        if 'crs' in value and value['crs'].get('type', None) == 'name':
            name = value['crs']['properties']['name']
            if name.startswith('urn:ogc:def:crs:EPSG::'):
                geom.srid = int(name.replace('urn:ogc:def:crs:EPSG::', ''))

        if geom.srid is None:
            geom.srid = 4326
        if geom.srid != srid:
            geom.transform(srid)
        return geom


class LabelRelatedField(serializers.RelatedField):
    def to_representation(self, obj):
        return str(obj)


# TODO: investigate use of SlugRelatedField and/or PrimaryKeyRelatedField
class IDRelatedField(serializers.RelatedField):
    def to_representation(self, obj):
        return get_object_id(obj)

    def to_internal_value(self, data):
        return get_by_identifier(self.queryset, data)


class ContentTypeField(serializers.SlugRelatedField):
    queryset = ContentType.objects.all()
    slug_field = "model"

    def __init__(self, **kwargs):
        super(serializers.RelatedField, self).__init__(**kwargs)


class LocalDateTimeField(serializers.ReadOnlyField):
    def to_representation(self, value):
        if value is None:
            return None
        if value.tzinfo:
            value = timezone.localtime(value)
        return value.strftime('%Y-%m-%d %I:%M %p')


class ModelSerializer(serializers.ModelSerializer):
    serializer_related_field = IDRelatedField
    add_label_fields = True

    def __init__(self, *args, **kwargs):
        for field in ('Geometry', 'GeometryCollection',
                      'Point', 'LineString', 'Polygon',
                      'MultiPoint', 'MultiLineString', 'MultiPolygon'):
            self.serializer_field_mapping[
                getattr(model_fields, field + 'Field')
            ] = GeometryField
        super(ModelSerializer, self).__init__(*args, **kwargs)

    @property
    def router(self):
        if 'view' not in self.context and 'router' not in self.context:
            return None
        if 'view' in self.context:
            return self.context['view'].router
        else:
            return self.context['router']

    @property
    def is_detail(self):
        if getattr(self.Meta, 'depth', 0) > 0:
            return True
        view = self.context.get('view', None)
        return view and view.action != "list"

    @property
    def is_edit(self):
        view = self.context.get('view', None)
        return view and view.action == 'edit'

    @property
    def is_geojson(self):
        request = self.context.get('request', None)
        if not request:
            return
        if request.accepted_renderer.format == 'geojson':
            return True
        return False

    def get_fields(self, *args, **kwargs):
        fields = super(ModelSerializer, self).get_fields(*args, **kwargs)
        fields = self.update_id_fields(fields)
        fields.update(self.get_label_fields(fields))
        if not self.is_detail:
            for field in getattr(self.Meta, 'list_exclude', []):
                fields.pop(field, None)
        return fields

    def update_id_fields(self, fields):
        if 'id' not in self.fields:
            conf = get_ct(self.Meta.model).get_config() or {}
            lookup = conf.get('lookup', None)
            if lookup and lookup != 'id':
                fields['id'] = serializers.ReadOnlyField(source=lookup)

        info = model_meta.get_field_info(self.Meta.model)

        # In list views, remove [fieldname] as an attribute in favor of
        # [fieldname]_id.
        for name, field in info.forward_relations.items():
            include = getattr(self.Meta, "fields", [])
            exclude = getattr(self.Meta, "exclude", [])
            if name in exclude or (include and name not in include):
                fields.pop(name, None)
                continue

            id_field, id_field_kwargs = self.build_relational_field(
                name, field
            )
            id_field_kwargs['source'] = name
            fields[name + '_id'] = id_field(**id_field_kwargs)

            if name in fields:
                # Update/remove DRF default foreign key field (w/o _id)
                if self.is_detail and isinstance(
                        fields[name], serializers.Serializer):
                    # Nested representation, keep for detail template context
                    fields[name].read_only = True
                else:
                    # Otherwise we don't need this field
                    del fields[name]

        return fields

    def get_label_fields(self, default_fields):
        if not self.add_label_fields:
            return {}
        fields = {}

        exclude = getattr(self.Meta, 'exclude', [])
        if 'label' not in exclude and 'label' not in default_fields:
            fields['label'] = serializers.ReadOnlyField(source='__str__')

        info = model_meta.get_field_info(self.Meta.model)

        # Add labels for dates and fields with choices
        for name, field in info.fields.items():
            if isinstance(field, model_fields.DateTimeField):
                fields[name + '_label'] = LocalDateTimeField(source=name)
            if field.choices:
                fields[name + '_label'] = serializers.ReadOnlyField(
                    source='get_%s_display' % name
                )

        # Add labels for related fields
        for name, field in info.forward_relations.items():
            if name in getattr(self.Meta, "exclude", []):
                continue

            f, field_kwargs = self.build_relational_field(
                name, info.forward_relations[name],
            )
            label_field_kwargs = {
                'source': name,
                'read_only': True,
            }
            if field_kwargs.get('many', None):
                label_field_kwargs['many'] = field_kwargs['many']
            fields[name + '_label'] = LabelRelatedField(**label_field_kwargs)

        return fields

    @classmethod
    def for_model(cls, model_class):
        class Serializer(cls):
            class Meta(getattr(cls, 'Meta', object)):
                model = model_class
        return Serializer

    @classmethod
    def for_depth(cls, serializer_depth):
        class Serializer(cls):
            class Meta(getattr(cls, 'Meta', object)):
                depth = serializer_depth
        return Serializer

    def build_nested_field(self, field_name, relation_info, nested_depth):
        field_class, field_kwargs = super(
            ModelSerializer, self
        ).build_nested_field(
            field_name, relation_info, nested_depth
        )
        field_kwargs['required'] = False
        if self.router:
            field_class = self.router.get_serializer_for_model(
                relation_info.related_model, nested_depth
            )
        return field_class, field_kwargs

    def to_internal_value(self, data):
        if html.is_html_input(data):
            data = html.parse_json_form(data)
        return super(ModelSerializer, self).to_internal_value(data)



from rest_framework import serializers


class ChartModelSerializer(serializers.ModelSerializer):
    class Meta:
        pandas_index = ["date"]
        pandas_unstacked_header = ["series", "units", "parameter"]

        pandas_scatter_coord = ["units", "parameter"]
        pandas_scatter_header = ["series"]

        pandas_boxplot_group = "series"
        pandas_boxplot_date = "date"
        pandas_boxplot_header = ["units", "parameter"]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.contrib.gis.db.models.fields

from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Location',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False, verbose_name='ID', auto_created=True)),
                ('name', models.CharField(max_length=255, blank=True, null=True)),
                ('is_primary', models.BooleanField(default=False)),
                ('geometry', django.contrib.gis.db.models.fields.GeometryField(srid=settings.SRID)),
                ('accuracy', models.IntegerField(blank=True, null=True)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
            options={
                'db_table': 'wq_location',
            },
            bases=(models.Model,),
        ),
    ]



from wq.db import rest
from .models import Identifier, Authority
from .serializers import IdentifierSerializer

rest.router.register_model(Identifier, serializer=IdentifierSerializer)
rest.router.register_model(Authority)



from wq.db import rest
from .models import Relationship, InverseRelationship, RelationshipType
from .serializers import (
    RelationshipSerializer,
    InverseRelationshipSerializer,
    RelationshipTypeSerializer
)

rest.router.register_model(
    Relationship, serializer=RelationshipSerializer
)
rest.router.register_model(
    InverseRelationship, serializer=InverseRelationshipSerializer
)
rest.router.register_model(
    RelationshipType, serializer=RelationshipTypeSerializer
)



from django.core.exceptions import ImproperlyConfigured

raise ImproperlyConfigured("""
wq.db.rest.settings has been renamed to wq.db.default_settings.
Please update your configuration.
See https://wq.io/wq.db/releases/v0.8.0 for more information.
""")



from rest_pandas import (
    PandasView, PandasUnstackedSerializer, PandasScatterSerializer,
    PandasBoxplotSerializer,
)
from wq.db.patterns.identify.filters import IdentifierFilterBackend
from .serializers import ChartModelSerializer


class ChartView(PandasView):
    serializer_class = ChartModelSerializer
    filter_backends = [IdentifierFilterBackend]


class TimeSeriesMixin(object):
    """
    For use with chart.timeSeries() in wq/chart.js
    """
    pandas_serializer_class = PandasUnstackedSerializer


class ScatterMixin(object):
    """
    For use with chart.scatter() in wq/chart.js
    """
    pandas_serializer_class = PandasScatterSerializer


class BoxPlotMixin(object):
    """
    For use with chart.boxplot() in wq/chart.js
    """
    pandas_serializer_class = PandasBoxplotSerializer



from wq.db import rest


def is_authenticated(request):
    return {
        'is_authenticated': request.user.is_authenticated()
    }


def social_auth(request):
    if (not request.user.is_authenticated()
            or not hasattr(request.user, 'social_auth')
            or request.user.social_auth.count() == 0):
        return {}

    return {
        'social_auth': {
            'accounts': [
                rest.router.serialize(auth)
                for auth in request.user.social_auth.all()
            ]
        }
    }



from wq.db import rest
from .views import LoginView, LogoutView, UserViewSet
from .serializers import UserSerializer, HAS_SOCIAL_AUTH
from django.contrib.auth.models import User

rest.router.add_page('login', {'url': 'login'}, LoginView)
rest.router.add_page('logout', {'url': 'logout'}, LogoutView)

rest.router.register_viewset(User, UserViewSet)
rest.router.register_serializer(User, UserSerializer)

if HAS_SOCIAL_AUTH:
    # Configure UserSocialAuth as nested serializer for User
    # (without actually registering it)
    from social.apps.django_app.default.models import UserSocialAuth
    from .serializers import SocialAuthSerializer
    rest.router.register_serializer(UserSocialAuth, SocialAuthSerializer)
    rest.router.register_config(UserSocialAuth, {'url': 'social_auth'})



# Use Django's template loader with Pystache's renderer
# NOTE: Support for multiple template engines was introduced in Django 1.8.
# This code includes hacks to support both 1.7 and 1.8.  When 1.9 comes out and
# we drop support for 1.7, we should be able to clean this up a bit.
from pystache.renderer import Renderer as PystacheRenderer
from django.template.loaders.filesystem import Loader as FileSystemLoader
from django.template.base import Template as DjangoTemplate
from django.utils.encoding import force_text
from django.utils.version import get_version

from django.utils.functional import empty
from django.utils.six import PY3
DJANGO18 = get_version() >= "1.8"


class Renderer(PystacheRenderer):
    def __init__(self, template, *args, **kwargs):
        self.template = template
        return super(Renderer, self).__init__(*args, **kwargs)

    def get_template_source(self, name):
        kwargs = {}
        if self.template.engine:
            kwargs['engine'] = self.template.engine
        template, origin = Loader(**kwargs).load_template(name)
        return template.source

    def _make_load_template(self):
        return self.get_template_source

    def _make_load_partial(self):
        def load_partial(name):
            return self.get_template_source('partials/' + name + '.html')
        return load_partial

    def _make_resolve_context(self):
        resolve_context = super(Renderer, self)._make_resolve_context()
        if DJANGO18 and PY3:
            return resolve_context

        def resolve(context, name):
            value = resolve_context(context, name)
            # Unwrap Django SimpleLazyObject
            if hasattr(value, '_wrapped'):
                if value._wrapped is empty:
                    value._setup()
                return value._wrapped
            return value
        return resolve

    def str_coerce(self, val):
        if val is None:
            return ""
        return force_text(val)


class Template(DjangoTemplate):
    def __init__(self, source, engine):
        # save raw source for actual rendering by Pystache
        self.source = source
        self.engine = engine

    def _render(self, context):
        renderer = Renderer(template=self)
        return renderer.render(self.source, *context)


class Loader(FileSystemLoader):
    def load_template(self, template_name, template_dirs=None):
        source, origin = self.load_template_source(
            template_name, template_dirs
        )
        template = Template(source, engine=getattr(self, 'engine', None))
        return template, origin
